
HealPointsCalculator = {
  HP_LastPowerfulCurrent = nil;
  HP_LastPowerfulNew = nil;
  HP_LastEfficientCurrent = nil;
  HP_LastEfficientNew = nil;
};

local BonusPercentInt;
local BonusPercentSpi;
local BonusPercentMana;

local Paladin_BoL =  {	{ HL = 210, FoL = 60},
						{ HL = 300, FoL = 85},
						{ HL = 400, FoL = 115},
						{ HL = 580, FoL = 185}};
            
-- Helper functions
local function getHPCast(spell, healing, spellCrit)
  local percentBonus = HealPointsSpells:getSpellBonus(spell);
  local healFactor, healFactor2 = HealPointsSpells:getHealingFactor(spell);
  spellCrit = spellCrit + HealPointsSpells:getSpellCritBonus(spell);
  healing = healing + HealPointsSpells:getHealingBonus(spell);
  
  local out = (spell['avg'] + healing * healFactor) * (1 + percentBonus);
  
	if (spell['type'] == 'regrowth') then
		-- Static part
		out = out * (1 + spellCrit / 200);		-- Crit adds 50% for non-hot
		-- HOT part
		out = out + (spell['hotavg'] + healing* healFactor2) * (1 + percentBonus);
    
	elseif (spell['type'] == 'lifebloom') then
		-- Static part
    out = (spell['avg'] + (healing + HealPointsBS:GetBonus('IMPLIFEBLOOMBURST')) * healFactor2) * (1 + percentBonus);
		out = out * (1 + spellCrit / 200);		-- Crit adds 50% for non-hot
		-- HOT part
		out = out + (spell['hotavg'] + (healing + HealPointsBS:GetBonus('IMPLIFEBLOOM')) * healFactor) * (1 + percentBonus);
    
  elseif (spell['type'] == 'chain') then
		local impChainHeal = HealPointsBS:GetBonus('IMPCHAINHEAL');
		out = out * (1 + 0.5 * (1 + impChainHeal / 100) + 0.25 * (1 + impChainHeal / 100)); -- Assumes that the spell hits all 3 targets
		out = out * (1 + (spellCrit / 200));												-- Crit adds 50% for non-hot

  elseif (spell['name'] == HPC_SPELL_HL or spell['name'] == HPC_SPELL_FOL) then
    if (getglobal("HealPointsOptionsConfigBoL"):GetChecked() == 1) then
			local bolRank = HealPointsSpells:getHighestSpellRank(HPC_SPELL_BOL);
			if (bolRank > 0) then
				local bol = Paladin_BoL[bolRank];
				out = out + (spell['bolfactor'] * (bol[spell['abbr']] + HealPointsBS:GetBonus('IMPBOL'))) * (1 + percentBonus);
			end
		end
		out = out * (1 + (spellCrit / 200));		-- Crit adds 50% for non-hot
  
  elseif (spell['name'] == HPC_SPELL_HW) then
		out = out * (1 + (spellCrit / 200));		-- Crit adds 50% for non-hot
		if (HealPointsBS:GetBonus('JUMPHEALINGWAVE') > 0) then
			local factor = 1 - HealPointsBS:GetBonus('JUMPHEALINGWAVE') / 100;
			out = out * (1 + factor + factor * factor);
		end
  
  elseif (spell['name'] == HPC_SPELL_GH) then  
		out = out * (1 + (spellCrit / 200));		-- Crit adds 50% for non-hot
		if (HealPointsBS:GetBonus('GHEALRENEW') > 0) then
			local renew = HealPointsSpells:getSpell(HPC_SPELL_RENEW, HealPointsBS:GetBonus('GHEALRENEW'));
      local renewHealFactor = HealPointsSpells:getHealingFactor(renew);
      healing = healing - HealPointsSpells:getHealingBonus(spell);
			out = out + renew['avg'] + healing * renewHealFactor; -- Assumes renew effect isn't overwritten
		end
  
  elseif (spell['type'] == 'normal' or spell['type'] == 'instant') then
		out = out * (1 + (spellCrit / 200));		-- Crit adds 50% for non-hot
  end

	return out;
end

local function getHPSec(spell, healing, spellCrit)
	local hpCast = getHPCast(spell, healing, spellCrit);

	if (spell['name'] == HPC_SPELL_GH and HealPointsBS:GetBonus('GHEALRENEW') > 0) then
		local renew = HealPointsSpells:getSpell(HPC_SPELL_RENEW, HealPointsBS:GetBonus('GHEALRENEW'));
    local renewHealFactor = HealPointsSpells:getHealingFactor(renew);
    healing = healing + HealPointsSpells:getHealingBonus(renew);
		return (hpCast - renew['avg'] - healing * renewHealFactor) / HealPointsSpells:getCastTime(spell); -- Remove renew-effect from HP/s.
	elseif (spell['type'] == 'regrowth') then
		return hpCast / (HealPointsSpells:getCastTime(spell) + HealPointsSpells:getDuration(spell));
  elseif (spell['type'] == 'hot' or spell['type'] == 'lifebloom') then
    return hpCast / HealPointsSpells:getDuration(spell);
	else
		return hpCast / HealPointsSpells:getCastTime(spell);
	end
end

local function getHPMana(spell, healing, spellCrit)
	local hpcast = getHPCast(spell, healing, spellCrit);
	local manaCost = HealPointsSpells:getManaCost(spell, spellCrit);
	return hpcast / manaCost;
end

local function getHPTotalRegen(spell, healing, spellCrit, manaRegenCasting, mana) 
	local hpPerCast = getHPCast(spell, healing, spellCrit);
	local manaCost = HealPointsSpells:getManaCost(spell, spellCrit);

	local interval;
	if (spell['type'] == 'regrowth') then
		interval = HealPointsSpells:getCastTime(spell) + HealPointsSpells:getDuration(spell);
  elseif (spell['type'] == 'hot' or spell['type'] == 'lifebloom') then
    interval = HealPointsSpells:getDuration(spell);
  else
    interval = HealPointsSpells:getCastTime(spell);
	end
  
	local manaUsageSpell = manaCost - interval * (manaRegenCasting / 5);
  local hpPerMana = hpPerCast / manaUsageSpell;
	local hpPerManaBar = hpPerMana * mana;
	
	local numberCasts = mana / manaUsageSpell;
	local totalTime = numberCasts * interval;
  
	return hpPerManaBar, totalTime;
end

local function computeTotalHealingNormal(spell, healing, seconds, startMana, spellCrit, manaRegenCasting, manaRegenNormal)
	--DEFAULT_CHAT_FRAME:AddMessage("Computing total healing for "..spell['name'].."("..spell['rank']..")  secs="..seconds.."  mana="..startMana,1,1,1);
  local hps = getHPSec(spell, healing, spellCrit);
  local castTime = HealPointsSpells:getCastTime(spell);
  local manaCost = HealPointsSpells:getManaCost(spell, spellCrit);

  -- Phase 1: Has mana to chain-cast
  local manaUsageSecond = (manaCost / castTime) - (manaRegenCasting / 5);
  local secondsPhase1 = math.min(seconds, startMana / manaUsageSecond);
  local totalHealed = secondsPhase1 * hps;		

  -- Phase 2: Has no mana left
  local secondsPhase2 = seconds - secondsPhase1;
  if (secondsPhase2 > 0) then
    local cycleTime;                      -- Seconds between each casting
    if (manaRegenCasting > manaCost) then 	-- chaincast inside 5s rule
      local secs = manaCost / (manaRegenCasting / 5);
      cycleTime = math.max(secs, castTime);
    else
      local manaToRegen = manaCost - manaRegenCasting;
      cycleTime = 5 + manaToRegen / (manaRegenNormal / 5);
    end
    local hpsPhase2 = hps * castTime / cycleTime;
    local healedPhase2 = hpsPhase2 * secondsPhase2;
    totalHealed = totalHealed + healedPhase2;
  end

  return totalHealed;
end

local function computeTotalHealingTrans8(spell, healing, seconds, startMana, spellCrit, manaRegenCasting, manaRegenNormal)
  --DEFAULT_CHAT_FRAME:AddMessage("Computing total healing for "..spell['name'].."("..spell['rank']..")  secs="..seconds.."  mana="..startMana,1,1,1);
  local hps = getHPSec(spell, healing, spellCrit);
  local manaCost = HealPointsSpells:getManaCost(spell, spellCrit);
  local castTime = HealPointsSpells:getCastTime(spell);
  local numtargets = HealPoints.db.char.hot['numtargets'];

  local renew = HealPointsSpells:getSpell(HPC_SPELL_RENEW, HealPointsBS:GetBonus('GHEALRENEW'));
  local renewHealFactor = HealPointsSpells:getHealingFactor(renew);
  healing = healing + HealPointsSpells:getHealingBonus(renew);

  -- Phase 1: Has mana to chain-cast
  local manaUsageSecond = (manaCost / castTime) - (manaRegenCasting / 5);
  local secondsPhase1 = math.min(seconds, startMana / manaUsageSecond);
  local secsBetweenSameTarget = numtargets * castTime;
  if (numtargets == 0) then
    secsBetweenSameTarget = castTime;
  end
  local ticsPerCast = secsBetweenSameTarget / 3;
  ticsPerCast = math.min(ticsPerCast, renew['duration'] / 3);
  local spellsCast = secondsPhase1 / castTime;

  local burstHealed = secondsPhase1 * hps;
  local hotHealed = spellsCast * ticsPerCast * (renew['avg'] + healing * renewHealFactor) * 3 / renew['duration'];
  local totalHealed = burstHealed + hotHealed;
  
  -- Phase 2: Has no mana left
  local secondsPhase2 = seconds - secondsPhase1;
  if (secondsPhase2 > 0) then
    local cycleTime;                      -- Seconds between each casting
    if (manaRegenCasting > manaCost) then 	-- chaincast inside 5s rule
      local secs = manaCost / (manaRegenCasting / 5);
      cycleTime = math.max(secs, castTime);
    else
      local manaToRegen = manaCost - manaRegenCasting;
      cycleTime = 5 + manaToRegen / (manaRegenNormal / 5);
    end
    secsBetweenSameTarget = numtargets * cycleTime;
    if (numtargets == 0) then
      secsBetweenSameTarget = cycleTime;
    end
    ticsPerCast = secsBetweenSameTarget / 3;
    ticsPerCast = math.min(ticsPerCast, renew['duration'] / 3);
    spellsCast = secondsPhase2 / cycleTime;
    burstHealed = secondsPhase2 * hps * castTime / cycleTime;
    hotHealed = spellsCast * ticsPerCast * (renew['avg'] + healing * renewHealFactor) * 3 / renew['duration'];
    local healedPhase2 = burstHealed + hotHealed;
    totalHealed = totalHealed + healedPhase2;
  end

  return totalHealed;
end

local function computeTotalHealingRegrowth(spell, healing, seconds, startMana, spellCrit, manaRegenCasting, manaRegenNormal)
	--DEFAULT_CHAT_FRAME:AddMessage("Computing total healing for "..spell['name'].."("..spell['rank']..")  secs="..seconds.."  mana="..startMana,1,1,1);
  local manaCost = HealPointsSpells:getManaCost(spell, spellCrit);
  local castTime = HealPointsSpells:getCastTime(spell);
  local duration = HealPointsSpells:getDuration(spell);
  local percentBonus = HealPointsSpells:getSpellBonus(spell);
  local healFactorBurst, healFactorHot = HealPointsSpells:getHealingFactor(spell);
  local numtargets = HealPoints.db.char.hot['numtargets'];
  spellCrit = spellCrit + HealPointsSpells:getSpellCritBonus(spell);
  healing = healing + HealPointsSpells:getHealingBonus(spell);
  
  local avgBurst = spell['avg'] * (1 + percentBonus) + healing * healFactorBurst;
  local avgHot = spell['hotavg'] * (1 + percentBonus) + healing * healFactorHot;

  -- Phase 1: Has mana to chain-cast
  local manaUsageSecond = (manaCost / castTime) - (manaRegenCasting / 5);
  local secondsPhase1 = math.min(seconds, startMana / manaUsageSecond);
  local secsBetweenSameTarget = numtargets * castTime;
  if (numtargets == 0) then
    secsBetweenSameTarget = castTime;
  end
  local ticsPerCast = secsBetweenSameTarget / 3;
  ticsPerCast = math.min(ticsPerCast, duration / 3);
  local spellsCast = secondsPhase1 / castTime;

  local burstHealed = spellsCast * avgBurst;
  burstHealed = burstHealed + (spellCrit / 200) * burstHealed;		-- Crit adds 50% for non-hot
  local hotHealed = spellsCast * ticsPerCast * avgHot * 3 / duration;
  
  local totalHealed = burstHealed + hotHealed;
  
  -- Phase 2: Has no mana left
  local secondsPhase2 = seconds - secondsPhase1;
  if (secondsPhase2 > 0) then
    local cycleTime;                      -- Seconds between each casting
    if (manaRegenCasting > manaCost) then 	-- chaincast inside 5s rule
      local secs = manaCost / (manaRegenCasting / 5);
      cycleTime = math.max(secs, castTime);
    else
      local manaToRegen = manaCost - manaRegenCasting;
      cycleTime = 5 + manaToRegen / (manaRegenNormal / 5);
    end
    secsBetweenSameTarget = numtargets * cycleTime;
    if (numtargets == 0) then
      secsBetweenSameTarget = cycleTime;
    end
    ticsPerCast = secsBetweenSameTarget / 3;
    ticsPerCast = math.min(ticsPerCast, duration / 3);
    spellsCast = secondsPhase2 / cycleTime;
    burstHealed = spellsCast * avgBurst;
    burstHealed = burstHealed + (spellCrit / 200) * burstHealed;		-- Crit adds 50% for non-hot
    hotHealed = spellsCast * ticsPerCast * avgHot * 3 / duration;
    local healedPhase2 = burstHealed + hotHealed;
    totalHealed = totalHealed + healedPhase2;
  end

  return totalHealed;
end

local function computeTotalHealingLifebloom(spell, healing, seconds, startMana, spellCrit, manaRegenCasting, manaRegenNormal)
  local hpCast = getHPCast(spell, healing, spellCrit); 
  local duration = HealPointsSpells:getDuration(spell);
  local manaCost = HealPointsSpells:getManaCost(spell, spellCrit); 
  local numtargets = HealPoints.db.char.hot['numtargets'];

	-- Phase 1: Has enough mana/regen to keep lifebloom up on numtargets constantly
  local cycleTime = math.max(duration, numtargets * 1.5); -- time between casts on the same target
  local manaUsageCycle = numtargets * manaCost;
  local regenTimeCasting = math.min(cycleTime, 5 + (numtargets - 1) * 1.5);
  local regenTimeNormal = math.max(0, cycleTime - regenTimeCasting);
  local manaRegainCycle = regenTimeCasting * (manaRegenCasting / 5) + regenTimeNormal * (manaRegenNormal / 5);
  local healingPerCycle = numtargets * hpCast;
  local numCycles = seconds / cycleTime;
  if (manaRegainCycle < manaUsageCycle) then -- if you regen more than you use, you'll never go oom.
      numCycles = math.min(numCycles, startMana / (manaUsageCycle - manaRegainCycle));
  end
  local secondsPhase1 = numCycles * cycleTime;
  local totalHealed = numCycles * healingPerCycle;

	-- Phase 2: Has no mana left
	local secondsPhase2 = seconds - secondsPhase1;
	if (secondsPhase2 > 0) then
    local manaToRegen = manaUsageCycle - regenTimeCasting * (manaRegenCasting / 5);
		cycleTime = regenTimeCasting + manaToRegen / (manaRegenNormal / 5);                      
    numCycles = secondsPhase2 / cycleTime;
    local healedPhase2 = numCycles * healingPerCycle;
		totalHealed = totalHealed + healedPhase2;
	end

	return totalHealed;
end

local function computeTotalHealingHot(spell, healing, seconds, startMana, manaRegenCasting, manaRegenNormal)
  local hpCast = getHPCast(spell, healing, 0); -- Can't crit
  local duration = HealPointsSpells:getDuration(spell);
  local manaCost = HealPointsSpells:getManaCost(spell, 0); -- Can't crit
  local numtargets = HealPoints.db.char.hot['numtargets'];

	-- Phase 1: Has enough mana/regen to keep the HoT up on numtargets constantly
  local cycleTime = math.max(duration, numtargets * 1.5); -- time between casts on the same target
  local manaUsageCycle = numtargets * manaCost;
  local regenTimeCasting = math.min(cycleTime, 5 + (numtargets - 1) * 1.5);
  local regenTimeNormal = math.max(0, cycleTime - regenTimeCasting);
  local manaRegainCycle = regenTimeCasting * (manaRegenCasting / 5) + regenTimeNormal * (manaRegenNormal / 5);
  local healingPerCycle = numtargets * hpCast;
  local numCycles = seconds / cycleTime;
  if (manaRegainCycle < manaUsageCycle) then -- if you regen more than you use, you'll never go oom.
      numCycles = math.min(numCycles, startMana / (manaUsageCycle - manaRegainCycle));
  end
  local secondsPhase1 = numCycles * cycleTime;
  local totalHealed = numCycles * healingPerCycle;

	-- Phase 2: Has no mana left
	local secondsPhase2 = seconds - secondsPhase1;
	if (secondsPhase2 > 0) then
    local manaToRegen = manaUsageCycle - regenTimeCasting * (manaRegenCasting / 5);
		cycleTime = regenTimeCasting + manaToRegen / (manaRegenNormal / 5);                      
    numCycles = secondsPhase2 / cycleTime;
    local healedPhase2 = numCycles * healingPerCycle;
		totalHealed = totalHealed + healedPhase2;
	end

	return totalHealed;
end

local function computePowerPoints(healing, spellCrit, maxMana, manaRegenCasting, manaRegenNormal) 
  local mana = HealPoints.db.char.power['mana'] * maxMana / 100;
	if (HealPoints.db.char.power['auto'] == true) then
		local powerfulSpell;
    local maxPowerPoints = 0;
		for i = 1, 3 do 
			local spell = HealPointsSpells:getHighestKnownSpell(HealPointsSpells.SPELLTABLE[i]);
			if (spell ~= nil and (spell['type'] == 'normal' or spell['type'] == 'chain' or spell['type'] == 'hot' or spell['type'] == 'regrowth')) then
				local powerPoints;
				if (spell['type'] == 'hot') then
					powerPoints = computeTotalHealingHot(spell, healing, 60 * HealPoints.db.char.power['duration'], mana, manaRegenCasting, manaRegenNormal);
        elseif (spell['type'] == 'regrowth') then
					powerPoints = computeTotalHealingRegrowth(spell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, manaRegenCasting, manaRegenNormal);
        elseif (spell['name'] == HPC_SPELL_GH and HealPointsBS:GetBonus('GHEALRENEW') > 0) then
					powerPoints = computeTotalHealingTrans8(spell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, manaRegenCasting, manaRegenNormal);
        elseif (spell['name'] == HPC_SPELL_HT and HealPointsUtil:isPlayerBuffUp(HPC_BUFF_TREELI)) then
          powerPoints = 0;
				else
					powerPoints = computeTotalHealingNormal(spell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, manaRegenCasting, manaRegenNormal);
				end
				if (powerPoints > maxPowerPoints) then
					powerfulSpell = spell;
					maxPowerPoints = powerPoints;
				end
			end
		end
		return maxPowerPoints, powerfulSpell;

	else
		local powerfulSpell = HealPointsSpells:getSpell(HealPoints.db.char.power['spell'], HealPoints.db.char.power['rank']);
		local powerPoints;
		if (powerfulSpell['type'] == 'hot') then
			powerPoints = computeTotalHealingHot(powerfulSpell, healing, 60 * HealPoints.db.char.power['duration'], mana, manaRegenCasting, manaRegenNormal);
    elseif (powerfulSpell['type'] == 'regrowth') then
      powerPoints = computeTotalHealingRegrowth(powerfulSpell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, manaRegenCasting, manaRegenNormal);
    elseif (powerfulSpell['type'] == 'lifebloom') then
      powerPoints = computeTotalHealingLifebloom(powerfulSpell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, manaRegenCasting, manaRegenNormal);
    elseif (powerfulSpell['name'] == HPC_SPELL_GH and HealPointsBS:GetBonus('GHEALRENEW') > 0) then
      powerPoints = computeTotalHealingTrans8(powerfulSpell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, manaRegenCasting, manaRegenNormal);
		else
			powerPoints = computeTotalHealingNormal(powerfulSpell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, manaRegenCasting, manaRegenNormal);
		end
		return powerPoints, powerfulSpell;
	end	
end

local function computeEndurancePoints(healing, spellCrit, maxMana, manaRegenCasting, manaRegenNormal) 
  local mana = HealPoints.db.char.endurance['mana'] * maxMana / 100;
	if (HealPoints.db.char.endurance['auto'] == true) then
		local efficientSpell;
    local maxEndurancePoints = 0;  
		for i = 1, 3, 1 do
			for j = 1, table.getn(HealPointsSpells.SPELLTABLE[i]), 1 do
				local spell = HealPointsSpells.SPELLTABLE[i][j];
				if (spell ~= nil and (spell['type'] == 'normal' or spell['type'] == 'chain' or spell['type'] == 'hot' or spell['type'] == 'regrowth' or spell['type'] == 'lifebloom') and HealPointsSpells:getHighestSpellRank(spell['name']) >= spell['rank']) then
					local endurancePoints;
					if (spell['type'] == 'hot') then
						endurancePoints = computeTotalHealingHot(spell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, manaRegenCasting, manaRegenNormal);
          elseif (spell['type'] == 'regrowth') then
            endurancePoints = computeTotalHealingRegrowth(spell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, manaRegenCasting, manaRegenNormal);
          elseif (spell['type'] == 'lifebloom') then
            endurancePoints = computeTotalHealingLifebloom(spell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, manaRegenCasting, manaRegenNormal);          
          elseif (spell['name'] == HPC_SPELL_GH and HealPointsBS:GetBonus('GHEALRENEW') > 0) then
            endurancePoints = computeTotalHealingTrans8(spell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, manaRegenCasting, manaRegenNormal);        
          elseif (spell['name'] == HPC_SPELL_HT and HealPointsUtil:isPlayerBuffUp(HPC_BUFF_TREELI)) then
            endurancePoints = 0;
					else
						endurancePoints = computeTotalHealingNormal(spell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, manaRegenCasting, manaRegenNormal);
					end
					if (endurancePoints > maxEndurancePoints) then
						efficientSpell = spell;
						maxEndurancePoints = endurancePoints;
					end			
				end
			end
		end
		return maxEndurancePoints, efficientSpell;

	else
		local efficientSpell = HealPointsSpells:getSpell(HealPoints.db.char.endurance['spell'], HealPoints.db.char.endurance['rank']);
		local endurancePoints;
		if (efficientSpell['type'] == 'hot') then
			endurancePoints = computeTotalHealingHot(efficientSpell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, manaRegenCasting, manaRegenNormal);
    elseif (efficientSpell['type'] == 'regrowth') then
      endurancePoints = computeTotalHealingRegrowth(efficientSpell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, manaRegenCasting, manaRegenNormal);
    elseif (efficientSpell['type'] == 'lifebloom') then
      endurancePoints = computeTotalHealingLifebloom(efficientSpell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, manaRegenCasting, manaRegenNormal);
    elseif (efficientSpell['name'] == HPC_SPELL_GH and HealPointsBS:GetBonus('GHEALRENEW') > 0) then
      endurancePoints = computeTotalHealingTrans8(efficientSpell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, manaRegenCasting, manaRegenNormal);        
		else
			endurancePoints = computeTotalHealingNormal(efficientSpell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, manaRegenCasting, manaRegenNormal);
		end
		return endurancePoints, efficientSpell;
	end
end

local function computeHealpoints(healing, spellCrit, maxMana, manaRegenCasting, manaRegenNormal)
	local powerPoints, powerfulSpell = computePowerPoints(healing, spellCrit, maxMana, manaRegenCasting, manaRegenNormal);
	local endurancePoints, efficientSpell = computeEndurancePoints(healing, spellCrit, maxMana, manaRegenCasting, manaRegenNormal);
  local healPoints = powerPoints + endurancePoints;
	return healPoints, powerPoints, endurancePoints, powerfulSpell, efficientSpell;
end

-- Compute old and new values
local function getCurrentAndNewManaRegen(intellect, spirit, itemCastingRegen)
	local function getRegenSpirit(spi)																												
		local baseSpirit = min(50, spi);
		local moreSpirit = spi - baseSpirit;
		local baseRate = 0.25;
		local moreRate;
    local _, unitClass = UnitClass("player");	
		if (unitClass == "PRIEST" ) then
			moreRate = 0.125;
		elseif ( unitClass == "DRUID" ) then
			moreRate = 0.1125;
		else
			moreRate = 0.1; -- paladin, shaman
		end
		return 5 * (baseSpirit * baseRate + moreSpirit * moreRate);
	end

  local function getRegenInt(int)
    local mana5Int = 0;
    local talentDreamstate = HealPointsUtil:getTalentRank(HPC_TALENT_DREAMS);  -- 4/7/10% of intellect as mana/5
    local talentUnrelentStorm = HealPointsUtil:getTalentRank(HPC_TALENT_UNRSTO); -- 2% of intellect as mana/5 per rank
    if (talentDreamstate == 1) then
      mana5Int = 0.04 * int;
    elseif (talentDreamstate == 2) then
      mana5Int = 0.07 * int;
    elseif (talentDreamstate == 3) then
      mana5Int = 0.1 * int;
    end
    mana5Int = mana5Int + 0.02 * talentUnrelentStorm * int;
    return mana5Int;
  end
  
	local talentMeditation = HealPointsUtil:getTalentRank(HPC_TALENT_MEDITA);	-- 5% spi mana regen while casting per rank (priest)
	local talentIntensity = HealPointsUtil:getTalentRank(HPC_TALENT_INTENS); 	-- 5% spi mana regen while casting per rank (druid)
	local percentRegen = 0.05 * talentMeditation + 0.05 * talentIntensity + HealPointsBS:GetBonus('CASTINGREG') / 100;

	local manaRegenSpirit = getRegenSpirit(spirit);
	local manaRegen5 = itemCastingRegen;
  
	if (HealPointsUtil:isPlayerBuffIconUp("INV_Potion_45")) then -- Mageblood potion
		manaRegen5 = manaRegen5 + 12;
	end
	if (HealPointsUtil:isPlayerBuffIconUp("Spell_Nature_ManaRegenTotem")) then -- Nightfin soup
		manaRegen5 = manaRegen5 + 8;
	end
	
  local manaRegenCastingInt = getRegenInt(intellect);
  
	local manaRegenNormal = manaRegenSpirit + manaRegen5 + manaRegenCastingInt; 
	local manaRegenCasting = manaRegen5 + manaRegenCastingInt + percentRegen * manaRegenSpirit;

	local intellectNew = intellect + getglobal("HealPointsVariables1InputBox"):GetNumber() * (1 + BonusPercentInt);
	local spiritNew = spirit + getglobal("HealPointsVariables2InputBox"):GetNumber() * (1 + BonusPercentSpi);
	local percentRegenNew = percentRegen + getglobal("HealPointsVariables6InputBox"):GetNumber() / 100;
	local manaRegen5New = manaRegen5 + getglobal("HealPointsVariables4InputBox"):GetNumber();
	local manaRegenSpiritNew = getRegenSpirit(spiritNew);
  local manaRegenCastingIntNew = getRegenInt(intellectNew);

	local manaRegenNormalNew = manaRegenSpiritNew + manaRegen5New + manaRegenCastingIntNew + getglobal("HealPointsVariables5InputBox"):GetNumber();
	local manaRegenCastingNew = manaRegen5New + manaRegenCastingIntNew + percentRegenNew * (manaRegenSpiritNew + getglobal("HealPointsVariables5InputBox"):GetNumber());
	return manaRegenNormal, manaRegenCasting, manaRegenNormalNew, manaRegenCastingNew, percentRegen;
end

local function getCurrentAndNewMaxMana(mana, itemIntDelta)
	local currentMana = mana + (15 * itemIntDelta) * (1 + BonusPercentMana);
	
	local intDelta = getglobal("HealPointsVariables1InputBox"):GetNumber() * (1 + BonusPercentInt);
	local newMana = currentMana + (getglobal("HealPointsVariables3InputBox"):GetNumber() + 15 * intDelta) * (1 + BonusPercentMana);
	return currentMana, newMana;
end

local function getHealingDelta(itemIntDelta, itemSpiDelta, itemStrDelta)
  local holyGuidance = HealPointsUtil:getTalentRank(HPC_TALENT_HOLYGU);
  local lunarGuidance = HealPointsUtil:getTalentRank(HPC_TALENT_LUNGUI);
  local natureBlessing = HealPointsUtil:getTalentRank(HPC_TALENT_NATBLE);
	local spiritualGuidance = HealPointsUtil:getTalentRank(HPC_TALENT_SPIGUI);
  local nurturingInstinct = HealPointsUtil:getTalentRank(HPC_TALENT_NURINS);

  local healingDelta = (holyGuidance * 0.07 + lunarGuidance * 0.0833 + natureBlessing * 0.1 + HealPointsBS:GetBonus('HEALFROMINT')) * itemIntDelta;
  healingDelta = healingDelta + spiritualGuidance * 0.05 * itemSpiDelta;
  healingDelta = healingDelta + nurturingInstinct * 0.25 * itemStrDelta;
  if (HealPointsUtil:isPlayerBuffUp(HPC_BUFF_TREELI)) then
    healingDelta = healingDelta + 0.25 * itemSpiDelta;
  end
  
  local healingDeltaNew = getglobal("HealPointsVariables7InputBox"):GetNumber();
  itemIntDelta = itemIntDelta + getglobal("HealPointsVariables1InputBox"):GetNumber() * (1 + BonusPercentInt);
  itemSpiDelta = itemSpiDelta + getglobal("HealPointsVariables2InputBox"):GetNumber() * (1 + BonusPercentSpi);
  healingDeltaNew = healingDeltaNew + (holyGuidance * 0.07 + lunarGuidance * 0.0833 + natureBlessing * 0.1) * itemIntDelta;
  healingDeltaNew = healingDeltaNew + spiritualGuidance * 0.05 * itemSpiDelta;
  healingDeltaNew = healingDeltaNew + nurturingInstinct * 0.25 * itemStrDelta;
  if (HealPointsUtil:isPlayerBuffUp(HPC_BUFF_TREELI)) then
    healingDeltaNew = healingDeltaNew + 0.25 * itemSpiDelta;
  end
  
	return healingDelta, healingDeltaNew;
end

local function getCurrentAndNewSpellCrit(itemIntDelta, itemCritRatingDelta) 
  local function ratingToPercent(critRating)
    if (UnitLevel("player") < 61) then
      return critRating * 26 / (7 * math.max(10, UnitLevel("player")) - 56);
    else
      return critRating * (262 - 3 * UnitLevel("player")) / 1148;
    end  
  end

  -- BaseCrit: 3.33 (paladin), 0.8 (priest), 1.85 (druid), 3.7 (shaman)
  
	local intPerCrit = 0;
  local spellCrit = 0;

	local _, className = UnitClass("player");
	if (className == "PALADIN") then
--		intPerCrit = 54; @ lvl 60
		intPerCrit = 80; 
    spellCrit = GetSpellCritChance(2); -- Holy
	elseif (className == "PRIEST") then
--		intPerCrit = 59.5; @ lvl 60
		intPerCrit = 80;
    spellCrit = GetSpellCritChance(2); -- Holy
	elseif (className == "DRUID") then
--		intPerCrit = 61; @ lvl 60
		intPerCrit = 80; 
    spellCrit = GetSpellCritChance(4); -- Nature
	elseif (className == "SHAMAN") then
--		intPerCrit = 57.1429; @lvl 60
		intPerCrit = 80;
    spellCrit = GetSpellCritChance(4); -- Nature
	end

  local critFromDeltaRating = ratingToPercent(itemCritRatingDelta); 
  local critFromDeltaInt = itemIntDelta / intPerCrit;
  local currentSpellCrit = spellCrit + critFromDeltaRating + critFromDeltaInt;

	local critRatingDelta = itemCritRatingDelta + getglobal("HealPointsVariables8InputBox"):GetNumber();
	local intDelta = itemIntDelta + getglobal("HealPointsVariables1InputBox"):GetNumber() * (1 + BonusPercentInt);
  
  critFromDeltaRating = ratingToPercent(critRatingDelta); 
  critFromDeltaInt = intDelta / intPerCrit;
  local newSpellCrit = spellCrit + critFromDeltaRating + critFromDeltaInt;
  
	return currentSpellCrit, newSpellCrit;
end

-- Global functions
function HealPointsCalculator:updateStats() 		-- Update stats (int, spi, regen, ...)																				
  if (BonusPercentInt == nil) then
    self:talentsChanged();
  end

	-- Intellect
	local _, currentIntellect, _, _ = UnitStat("player", 4);
	local intDelta = getglobal("HealPointsVariables1InputBox"):GetNumber();
	local extraInt = BonusPercentInt * intDelta;
	HealPointsCalculatorUI:setStatInfo(1, currentIntellect, extraInt, "%5.0f");
	intDelta = intDelta + extraInt;
	
	-- Spirit
	local _, currentSpirit, _, _ = UnitStat("player", 5);
	local extraSpi = BonusPercentSpi * getglobal("HealPointsVariables2InputBox"):GetNumber();
	HealPointsCalculatorUI:setStatInfo(2, currentSpirit, extraSpi, "%5.0f");

	-- Max mana															
	local currentMana = UnitManaMax("player");
	local extraMana = BonusPercentMana * (getglobal("HealPointsVariables3InputBox"):GetNumber() + 15 * intDelta);
	HealPointsCalculatorUI:setStatInfo(3, currentMana, 15 * intDelta + extraMana, "%5.0f");
	
	-- Mana regen (casting, normal, %casting)
	local manaRegenNormal, manaRegenCasting, manaRegenNormalNew, manaRegenCastingNew, percentRegen = 
		getCurrentAndNewManaRegen(currentIntellect, currentSpirit, HealPointsBS:GetBonus('MANAREG'));
	local manaRegenCastingDelta = manaRegenCastingNew - manaRegenCasting - getglobal("HealPointsVariables4InputBox"):GetNumber();
	local manaRegenNormalDelta = manaRegenNormalNew - manaRegenNormal - getglobal("HealPointsVariables5InputBox"):GetNumber();
	
	HealPointsCalculatorUI:setStatInfo(4, manaRegenCasting, manaRegenCastingDelta, "%5.0f");
	HealPointsCalculatorUI:setStatInfo(5, manaRegenNormal, manaRegenNormalDelta, "%5.0f");
	HealPointsCalculatorUI:setStatInfo(6, 100 * percentRegen, 0, "%5.0f");
	
	-- +Healing
	local currentHealing = GetSpellBonusHealing();
  local _, healingDelta = getHealingDelta(0, 0, 0);
	HealPointsCalculatorUI:setStatInfo(7, currentHealing, healingDelta - getglobal("HealPointsVariables7InputBox"):GetNumber(), "%5.0f");
	
	-- Spell crit rating
	HealPointsCalculatorUI:setStatInfo(8, HealPointsBS:GetBonus('SPELLCRITRATING'), 0, "%5.0f");
end

function HealPointsCalculator:getSpellStats(spell) -- Get derived stats about specified spell  
  if (BonusPercentInt == nil) then
    self:talentsChanged();
  end

  local _, currentIntellect, _, _ = UnitStat("player", 4);
	local _, currentSpirit, _, _ = UnitStat("player", 5);
	local _, manaRegenCasting, _, manaRegenCastingNew = getCurrentAndNewManaRegen(currentIntellect, currentSpirit, HealPointsBS:GetBonus('MANAREG'));
	local currentMana, newMana = getCurrentAndNewMaxMana(UnitManaMax("player"), 0);
	local currentHealing = GetSpellBonusHealing();
  local _, healingDelta = getHealingDelta(0, 0, 0);
  local newHealing = currentHealing + healingDelta;
	local currentSpellCrit, newSpellCrit = getCurrentAndNewSpellCrit(0, 0);	

  local oldHPSec = getHPSec(spell, currentHealing, currentSpellCrit);
  local newHPSec = getHPSec(spell, newHealing, newSpellCrit);
  local oldHPMana = getHPMana(spell, currentHealing, currentSpellCrit);
  local newHPMana = getHPMana(spell, newHealing, newSpellCrit);
  local oldHPCast = getHPCast(spell, currentHealing, currentSpellCrit);
  local newHPCast = getHPCast(spell, newHealing, newSpellCrit);
  local oldHPTotal = oldHPMana * currentMana;
  local newHPTotal = newHPMana * newMana;
  local oldHPTotalRegen, t1 = getHPTotalRegen(spell, currentHealing, currentSpellCrit, manaRegenCasting, currentMana);
  local newHPTotalRegen, t2 = getHPTotalRegen(spell, newHealing, newSpellCrit, manaRegenCastingNew, newMana);
  local oldCrit = currentSpellCrit + HealPointsSpells:getSpellCritBonus(spell);
  local newCrit = newSpellCrit + HealPointsSpells:getSpellCritBonus(spell);
  
  return oldHPSec, newHPSec, oldHPMana, newHPMana, oldHPCast, newHPCast, oldHPTotal, newHPTotal, oldHPTotalRegen, newHPTotalRegen, t1, t2, oldCrit, newCrit;  
end

function HealPointsCalculator:updateHealPoints() -- Update the healpoints statistic
  if (BonusPercentInt == nil) then
    self:talentsChanged();
  end

	local _, currentIntellect, _, _ = UnitStat("player", 4);
	local _, currentSpirit, _, _ = UnitStat("player", 5);
	local manaRegenNormal, manaRegenCasting, manaRegenNormalNew, manaRegenCastingNew = getCurrentAndNewManaRegen(currentIntellect, currentSpirit, HealPointsBS:GetBonus('MANAREG'));
	local currentMana, newMana = getCurrentAndNewMaxMana(UnitManaMax("player"), 0);
	local currentHealing = GetSpellBonusHealing();
  local _, healingDelta = getHealingDelta(0, 0, 0);
  local newHealing = currentHealing + healingDelta;
	local currentSpellCrit, newSpellCrit = getCurrentAndNewSpellCrit(0, 0);	
	
	local healPointsCurrent, healedPowerful, healedEfficient, powerfulCurrent, efficientCurrent =
		computeHealpoints(currentHealing, currentSpellCrit, currentMana, manaRegenCasting, manaRegenNormal);
	
	local healPointsNew, healedPowerfulNew, healedEfficientNew, powerfulNew, efficientNew =
		computeHealpoints(newHealing, newSpellCrit, newMana, manaRegenCastingNew, manaRegenNormalNew);
	
	self.HP_LastPowerfulCurrent = powerfulCurrent;
	self.HP_LastPowerfulNew = powerfulNew;
	self.HP_LastEfficientCurrent = efficientCurrent;
	self.HP_LastEfficientNew = efficientNew;
	
  if (HealPoints_CalcFrame:IsVisible()) then
    HealPointsCalculatorUI:setHealPoints(healedPowerful, healedPowerfulNew, healedEfficient, healedEfficientNew);
  end
  if (CharacterFrame:IsVisible()) then
    HealPointsCharUI:setStats(healPointsCurrent, healedPowerful, healedEfficient, powerfulCurrent, efficientCurrent);
  end
end

function HealPointsCalculator:computeHealpointsDiff(frame, itemLink) 
	local function lookup(table, index)
		if (table[index] == nil) then
			return 0;
		end
		return table[index];
	end

	local function computeHP(itemBonuses, slotBonuses)
		local _, currentIntellect = UnitStat("player", 4);
		local _, currentSpirit = UnitStat("player", 5);
		local intDelta = HealPointsUtil:round((lookup(itemBonuses, "INT") - lookup(slotBonuses, "INT")) * (1 + BonusPercentInt));
		local spiDelta = HealPointsUtil:round((lookup(itemBonuses, "SPI") - lookup(slotBonuses, "SPI")) * (1 + BonusPercentSpi));
    local strDelta = HealPointsUtil:round(lookup(itemBonuses, "STR") - lookup(slotBonuses, "STR"));
		local ratingDelta = lookup(itemBonuses, "SPELLCRITRATING") - lookup(slotBonuses, "SPELLCRITRATING");
		local itemMana = HealPointsUtil:round(UnitManaMax("player") - (lookup(slotBonuses, "MANA") - lookup(itemBonuses, "MANA")) * (1 + BonusPercentMana));
		local itemMana5 = HealPointsBS:GetBonus('MANAREG') - lookup(slotBonuses, "MANAREG") + lookup(itemBonuses, "MANAREG");
    local healDelta, healDeltaNew = getHealingDelta(intDelta, spiDelta, strDelta);

		local iHealing = HealPointsUtil:round(GetSpellBonusHealing() + healDelta - lookup(slotBonuses, "HEAL") + lookup(itemBonuses, "HEAL"));
    local iHealingNew = HealPointsUtil:round(GetSpellBonusHealing() +  healDeltaNew - lookup(slotBonuses, "HEAL") + lookup(itemBonuses, "HEAL"));
		local iSpellCrit, iSpellCritNew = getCurrentAndNewSpellCrit(intDelta, ratingDelta);
		local iMana, iManaNew = getCurrentAndNewMaxMana(itemMana, intDelta);
		local iManaRegenNormal, iManaRegenCasting, iManaRegenNormalNew, iManaRegenCastingNew = 
      getCurrentAndNewManaRegen(currentIntellect + intDelta, currentSpirit + spiDelta, itemMana5);

    if (HealPoints.db.char.debug ~= 0) then
      DEFAULT_CHAT_FRAME:AddMessage("Tooltip debug: +int:"..intDelta.. " +spi:"..spiDelta.." healing:"..iHealing.." mana/5:"..itemMana5.." +critRating:"..ratingDelta,0.6,0.6,1);
    end
		local iHealPoints = computeHealpoints(iHealing, iSpellCrit, iMana, iManaRegenCasting, iManaRegenNormal);
    local iHealPointsNew = iHealPoints;
    if (iHealingNew ~= iHealing or iSpellCritNew ~= iSpellCrit or iManaNew ~= iMana or iManaRegenCastingNew ~= iManaRegenCasting or iManaRegenNormalNew ~= iManaRegenNormal) then    
      iHealPointsNew = computeHealpoints(iHealingNew, iSpellCritNew, iManaNew, iManaRegenCastingNew, iManaRegenNormalNew);
    end
		return iHealPoints, iHealPointsNew
	end

  if (BonusPercentInt == nil) then
    self:talentsChanged();
  end

  local itemBonuses, slotBonuses, slotBonuses2 = HealPointsBS:getItemSlotBonuses(frame, itemLink);
  if (slotBonuses == nil) then
    return 0, 0;
  end

  if (HealPoints.db.char.debug ~= 0) then
    DEFAULT_CHAT_FRAME:AddMessage("Bonuses for: "..itemLink,1,1,1);
    for bonus, details in pairs(itemBonuses) do
      HealPoints:Print(bonus..": "..details);
    end
--    DEFAULT_CHAT_FRAME:AddMessage("Replaces these bonuses: ",1,1,1);
--    for bonus, details in pairs(slotBonuses) do
--      HealPoints:Print(bonus..": "..details);
--    end
  end

	-- Current healpoints
	local _, currentIntellect = UnitStat("player", 4);
	local _, currentSpirit = UnitStat("player", 5);

	local currentHealing = GetSpellBonusHealing();
  local _, healingDelta = getHealingDelta(0, 0, 0);
  local newHealing = currentHealing + healingDelta;
	local spellCrit, spellCritNew = getCurrentAndNewSpellCrit(0, 0);
	local mana, manaNew = getCurrentAndNewMaxMana(UnitManaMax("player"), 0)
	local manaRegenNormal, manaRegenCasting, manaRegenNormalNew, manaRegenCastingNew = 
		getCurrentAndNewManaRegen(currentIntellect, currentSpirit, HealPointsBS:GetBonus('MANAREG'));

	local healPoints = computeHealpoints(currentHealing, spellCrit, mana, manaRegenCasting, manaRegenNormal);
  local healPointsNew = healPoints;
  if (newHealing ~= currentHealing or spellCritNew ~= spellCrit or manaNew ~= mana or manaRegenCastingNew ~= manaRegenCasting or manaRegenNormalNew ~= manaRegenNormal) then  
    healPointsNew = computeHealpoints(newHealing, spellCritNew, manaNew, manaRegenCastingNew, manaRegenNormalNew);
  end

	-- Difference
	local iHealPoints, iHealPointsNew = computeHP(itemBonuses, slotBonuses);
	if (slotBonuses2) then
		local iHealPoints2, iHealPointsNew2 = computeHP(itemBonuses, slotBonuses2);
		return iHealPoints - healPoints, iHealPointsNew - healPointsNew, iHealPoints2 - healPoints, iHealPointsNew2 - healPointsNew;
	else
		return iHealPoints - healPoints, iHealPointsNew - healPointsNew;
	end
end

function HealPointsCalculator:talentsChanged()
	BonusPercentMana = 0;
	BonusPercentInt = 0;
  BonusPercentSpi = 0;

	local _, raceName = UnitRace("player");
	if (raceName == "Human") then
		BonusPercentSpi = 0.1;
	end

	local _, className = UnitClass("player");	
	if (className == "PALADIN") then
		local divineIntellect = HealPointsUtil:getTalentRank(HPC_TALENT_DIVINT);  -- Increases intellect by 2% per rank
		BonusPercentInt = 0.02 * divineIntellect;
		
	elseif (className == "PRIEST") then
		local mentalStrength = HealPointsUtil:getTalentRank(HPC_TALENT_MENSTR); -- Increases maximum mana by 2% per rank
		BonusPercentMana = 0.02 * mentalStrength;
    
    local enlightenment = HealPointsUtil:getTalentRank(HPC_TALENT_ENLIGH); -- Increases int/spi by 1% per rank
    BonusPercentInt = 0.01 * enlightenment;
    BonusPercentSpi =  BonusPercentSpi + 0.01 * enlightenment;
    
    local spiritOfRedemption = HealPointsUtil:getTalentRank(HPC_TALENT_SPIRED);  -- Increases spi by 5%
    BonusPercentSpi = BonusPercentSpi + 0.05 * spiritOfRedemption;
    
  elseif (className == "DRUID") then
		local heartOfTheWild = HealPointsUtil:getTalentRank(HPC_TALENT_HEOFWI); -- Increases int by 4% per rank
		BonusPercentInt = 0.04 * heartOfTheWild;
    
    local livingSpirit = HealPointsUtil:getTalentRank(HPC_TALENT_LIVSPI); -- Increases spi by 5% per rank
		BonusPercentSpi = 0.05 * livingSpirit;    
    
    local survivalFittest = HealPointsUtil:getTalentRank(HPC_TALENT_SURFIT); -- Increases all stats by 1% per rank
    BonusPercentInt = BonusPercentInt + 0.01 * survivalFittest;
    BonusPercentSpi = BonusPercentSpi + 0.01 * survivalFittest;

  elseif (className == "SHAMAN") then
		local ancestralKnowledge = HealPointsUtil:getTalentRank(HPC_TALENT_ANCKNO); -- Increases maximum mana by 1% per rank
		BonusPercentMana = 0.01 * ancestralKnowledge;
	end
end